clear; close all; clc
set(0,'DefaultFigureWindowStyle','docked')

%% Landmarks
% legyenek ismert megfeleltetsek
N = 15;
scale = 200;
map = scale * rand(N,2);

% Plot map
figure(1)
axis equal
hold on
box on

xlim([-0.2*scale-100,1.1*scale+100])
ylim([-0.1*scale-50,1.1*scale])
for i=1:N
    rectangle('Position',[map(i,1), map(i,2),5,5])
end

%% Paramterek
dt=0.5;         % idlps [s]
Nt=2000;         % idpontok szma
T=dt*Nt;
t=dt:dt:T;      % idpontok
v=7.5;          % sebessg [m/s]

%% Modellek
% Mozgsi modell

g=@(v,w,fi) [-v/w*sin(fi)+v/w*sin(fi+w*dt);
             v/w*cos(fi)-v/w*cos(fi+w*dt);
             w*dt];

G=@(v,w,fi)[1 0 -v/w*sin(fi)+v/w*sin(fi+w*dt);
            0 1 v/w*cos(fi)-v/w*cos(fi+w*dt);
            0 0 1];

F=@(v,w,fi) [G(v,w,fi) zeros(3,2*N);
            zeros(2*N,3) eye(2*N)];
        
Q = zeros(3+2*N);
Q(1,1) = 0.1;
Q(2,2) = 0.1;
Q(3,3) = 0.001;
Q = Q / 1;
        
% Mrsi modell
H1=@(delta,q) [-sqrt(q)*delta(1) -sqrt(q)*delta(2) 0;
                    delta(2) -delta(1) -q];

H2=@(delta,q) [sqrt(q)*delta(1) sqrt(q)*delta(2);
                    -delta(2) delta(1)];
H=@(delta,q,j) 1/q*[H1(delta,q),zeros(2,2*j-2),H2(delta,q),zeros(2,2*N-2*j)];

R = [0.01, 0;
     0, 0.001] / 1;
        
%% A robot tja
r=linspace(120,70,Nt);    % vltoz sugr
w=-v./r;                   % vltoz szgsebessg
robot(1).x = [0,scale/2-20,pi/2]';
robot(Nt).x = [];

plot(robot(1).x(1),robot(1).x(2))
for k = 2:Nt
    robot(k).x = g(v,w(k-1),robot(k-1).x(3)) + robot(k-1).x;
    plot(robot(k).x(1),robot(k).x(2),'ko')
end

%% Init filter
X = [robot(1).x; zeros(2*N,1)];
P = ones(3+2*N)* 1e10;
P(1:3,1:3) = 0;
ok = 0;
for k = 2:Nt
    k;
    X = X + [g(v,w(k-1),robot(k-1).x(3)); zeros(2*N,1)];
    X(3) = mod(X(3),2*pi);
    P = F(v,w(k),X(3)) * P + F(v,w(k),X(3))' + Q;
    
    plot(X(1),X(2),'rx')
    %pause(0.001)
    %continue
    % Mrsi adatok generlsa
    D=34*5;   % ez a szenzor hattv
    Z=[];
       for i=1:N
        delta=[map(i,1)-robot(k).x(1); map(i,2)-robot(k).x(2)];
        q=delta'*delta;
        z(1,i)=sqrt(q);
        z(2,i)=atan2(delta(2),delta(1))-robot(k).x(3);
        z(:,i) = z(:,i) + sqrtm(R)*randn(2,1);
        if sqrt(q)<D
            
            Z=[Z,[z(:,i);i]];
            map(i,:);
            %Z
            % inicializlni kell, ha mg nem volt
            if(X(3+2*i)==0)
                X(3+2*i-1)=X(1)+z(1,i)*cos(z(2,i)+X(3));
                X(3+2*i+0)=X(2)+z(1,i)*sin(z(2,i)+X(3));
                disp('cont')
                ok = 1;
                %pause
                continue
            end
            H(delta,q,i);

            S = H(delta,q,i) * P * H(delta,q,i)' + R;
            K = P * H(delta,q,i)' * inv( S );
            % zajt tesznk r
            %X=X+[chol(Q(1:3,1:3))*randn(3,1);zeros(2*N,1)]
             delta=[map(i,1)-X(1);map(i,2)-X(2)];
             q=delta'*delta;
             
             zz(1)=sqrt(q);
             zz(2)=atan2(delta(2),delta(1))-X(3);
            
            rez = z(:,i)-zz(:);
            rez(2) = mod(rez(2),2*pi);
            if rez(2) > pi
                rez(2) = rez(2)-2*pi;
            end
            if rez(2) < -pi
                rez(2) = rez(2)+2*pi;
            end
            rez;
            X = X+K*(rez);
            P = P-K*H(delta,q,i)*P;
            %pause
        end
        
       end
       
      plot(X(1),X(2),'rx')

      if ok
          %pause
      end
    
end

XX=reshape(X(4:end)',2,N)'
XX-map

for i=1:N
    rectangle('Position',[XX(i,1), XX(i,2),5,5],'Curvature',1,'FaceColor','red')
end

s=diag(diag(P));
cor = s^(-1/2)*P*s^(-1/2);
figure
subplot(1,2,1)
imagesc(cor)
subplot(1,2,2)
imagesc(inv(cor))

